@@ -1,7 +1,7 @@
 # The MIT License
 #
 # Copyright (c) 2008 Bob Farrell
-# Copyright (c) 2012-2020 Sebastian Ramacher
+# Copyright (c) 2012-2021 Sebastian Ramacher
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -21,31 +21,46 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 
+# To gradually migrate to mypy we aren't setting these globally yet
+# mypy: disallow_untyped_defs=True
+# mypy: disallow_untyped_calls=True
+
 """
 Module to handle command line argument parsing, for all front-ends.
 """
 
-import importlib.util
-import os
-import sys
 import argparse
+from typing import Tuple, List, Optional, NoReturn, Callable
+import code
+import curtsies
+import cwcwidth
+import greenlet
+import importlib.util
+import logging
+import os
+import pygments
+import requests
+import sys
+from pathlib import Path
 
 from . import __version__, __copyright__
-from .config import default_config_path, loadini, Struct
+from .config import default_config_path, Config
 from .translations import _
 
+logger = logging.getLogger(__name__)
+
 
 class ArgumentParserFailed(ValueError):
     """Raised by the RaisingOptionParser for a bogus commandline."""
 
 
 class RaisingArgumentParser(argparse.ArgumentParser):
-    def error(self, msg):
+    def error(self, msg: str) -> NoReturn:
         raise ArgumentParserFailed()
 
 
-def version_banner(base="bpython"):
-    return "{} version {} on top of Python {} {}".format(
+def version_banner(base: str = "bpython") -> str:
+    return _("{} version {} on top of Python {} {}").format(
         base,
         __version__,
         sys.version.split()[0],
@@ -53,11 +68,18 @@ def version_banner(base="bpython"):
     )
 
 
-def copyright_banner():
-    return "{} See AUTHORS.rst for details.".format(__copyright__)
+def copyright_banner() -> str:
+    return _("{} See AUTHORS.rst for details.").format(__copyright__)
 
 
-def parse(args, extras=None, ignore_stdin=False):
+Options = Tuple[str, str, Callable[[argparse._ArgumentGroup], None]]
+
+
+def parse(
+    args: Optional[List[str]],
+    extras: Optional[Options] = None,
+    ignore_stdin: bool = False,
+) -> Tuple[Config, argparse.Namespace, List[str]]:
     """Receive an argument list - if None, use sys.argv - parse all args and
     take appropriate action. Also receive optional extra argument: this should
     be a tuple of (title, description, callback)
@@ -73,9 +95,12 @@ def parse(args, extras=None, ignore_stdin=False):
 
     parse(
         ['-i', '-m', 'foo.py'],
-        ('Front end-specific options',
-        'A full description of what these options are for',
-        callback))
+        (
+            'Front end-specific options',
+            'A full description of what these options are for',
+            callback
+        ),
+    )
 
 
     Return a tuple of (config, options, exec_args) wherein "config" is the
@@ -98,6 +123,7 @@ def parse(args, extras=None, ignore_stdin=False):
     parser.add_argument(
         "--config",
         default=default_config_path(),
+        type=Path,
         help=_("Use CONFIG instead of default config file."),
     )
     parser.add_argument(
@@ -118,6 +144,18 @@ def parse(args, extras=None, ignore_stdin=False):
         action="store_true",
         help=_("Print version and exit."),
     )
+    parser.add_argument(
+        "--log-level",
+        "-l",
+        choices=("debug", "info", "warning", "error", "critical"),
+        default="error",
+        help=_("Set log level for logging"),
+    )
+    parser.add_argument(
+        "--log-output",
+        "-L",
+        help=_("Log output file"),
+    )
 
     if extras is not None:
         extras_group = parser.add_argument_group(extras[0], extras[1])
@@ -128,7 +166,7 @@ def parse(args, extras=None, ignore_stdin=False):
         "args",
         nargs=argparse.REMAINDER,
         help=_(
-            "File to extecute and additional arguments passed on to the executed script."
+            "File to execute and additional arguments passed on to the executed script."
         ),
     )
 
@@ -147,25 +185,62 @@ def parse(args, extras=None, ignore_stdin=False):
         # Just let Python handle this
         os.execv(sys.executable, [sys.executable] + args)
 
-    config = Struct()
-    loadini(config, options.config)
+    # Configure logging handler
+    bpython_logger = logging.getLogger("bpython")
+    curtsies_logger = logging.getLogger("curtsies")
+    bpython_logger.setLevel(options.log_level.upper())
+    curtsies_logger.setLevel(options.log_level.upper())
+    if options.log_output:
+        handler = logging.FileHandler(filename=options.log_output)
+        handler.setFormatter(
+            logging.Formatter(
+                "%(asctime)s: %(name)s: %(levelname)s: %(message)s"
+            )
+        )
+        bpython_logger.addHandler(handler)
+        curtsies_logger.addHandler(handler)
+        bpython_logger.propagate = curtsies_logger.propagate = False
+    else:
+        bpython_logger.addHandler(logging.NullHandler())
+        curtsies_logger.addHandler(logging.NullHandler())
 
-    return config, options, options.args
+    logger.info("Starting bpython %s", __version__)
+    logger.info("Python %s: %s", sys.executable, sys.version_info)
+    logger.info("curtsies: %s", curtsies.__version__)
+    logger.info("cwcwidth: %s", cwcwidth.__version__)
+    logger.info("greenlet: %s", greenlet.__version__)
+    logger.info("pygments: %s", pygments.__version__)  # type: ignore
+    logger.info("requests: %s", requests.__version__)
+    logger.info("environment:")
+    for key, value in sorted(os.environ.items()):
+        if key.startswith("LC") or key.startswith("LANG") or key == "TERM":
+            logger.info("%s: %s", key, value)
+
+    return Config(options.config), options, options.args
 
 
-def exec_code(interpreter, args):
+def exec_code(
+    interpreter: code.InteractiveInterpreter, args: List[str]
+) -> None:
     """
-    Helper to execute code in a given interpreter. args should be a [faked]
-    sys.argv
+    Helper to execute code in a given interpreter, e.g. to implement the behavior of python3 [-i] file.py
+
+    args should be a [faked] sys.argv.
     """
-    with open(args[0]) as sourcefile:
-        source = sourcefile.read()
+    try:
+        with open(args[0]) as sourcefile:
+            source = sourcefile.read()
+    except OSError as e:
+        # print an error and exit (if -i is specified the calling code will continue)
+        print(f"bpython: can't open file '{args[0]}: {e}", file=sys.stderr)
+        raise SystemExit(e.errno)
     old_argv, sys.argv = sys.argv, args
     sys.path.insert(0, os.path.abspath(os.path.dirname(args[0])))
-    spec = importlib.util.spec_from_loader("__console__", loader=None)
+    spec = importlib.util.spec_from_loader("__main__", loader=None)
+    assert spec
     mod = importlib.util.module_from_spec(spec)
-    sys.modules["__console__"] = mod
-    interpreter.locals = mod.__dict__
-    interpreter.locals["__file__"] = args[0]
+    sys.modules["__main__"] = mod
+    interpreter.locals.update(mod.__dict__)  # type: ignore  # TODO use a more specific type that has a .locals attribute
+    interpreter.locals["__file__"] = args[0]  # type: ignore  # TODO use a more specific type that has a .locals attribute
     interpreter.runsource(source, args[0], "exec")
     sys.argv = old_argv
