@@ -1,18 +1,21 @@
-# coding: utf8
-
-from __future__ import unicode_literals
 import itertools
 import os
 import pydoc
 import string
 import sys
-from contextlib import contextmanager
 
-from curtsies.formatstringarray import FormatStringTest, fsarray
+from contextlib import contextmanager
+from typing import cast
+from curtsies.formatstringarray import (
+    fsarray,
+    assertFSArraysEqual,
+    assertFSArraysEqualIgnoringFormatting,
+)
 from curtsies.fmtfuncs import cyan, bold, green, yellow, on_magenta, red
+from curtsies.window import CursorAwareWindow
+from unittest import mock
 
 from bpython.curtsiesfrontend.events import RefreshRequestEvent
-from bpython.test import mock
 from bpython import config, inspection
 from bpython.curtsiesfrontend.repl import BaseRepl
 from bpython.curtsiesfrontend import replpainter
@@ -24,8 +27,7 @@ from bpython.test import FixLanguageTestCase as TestCase, TEST_CONFIG
 
 
 def setup_config():
-    config_struct = config.Struct()
-    config.loadini(config_struct, TEST_CONFIG)
+    config_struct = config.Config(TEST_CONFIG)
     config_struct.cli_suggestion_width = 1
     return config_struct
 
@@ -33,7 +35,14 @@ def setup_config():
 class ClearEnviron(TestCase):
     @classmethod
     def setUpClass(cls):
-        cls.mock_environ = mock.patch.dict("os.environ", {}, clear=True)
+        cls.mock_environ = mock.patch.dict(
+            "os.environ",
+            {
+                "LC_ALL": os.environ.get("LC_ALL", "C.UTF-8"),
+                "LANG": os.environ.get("LANG", "C.UTF-8"),
+            },
+            clear=True,
+        )
         cls.mock_environ.start()
         TestCase.setUpClass()
 
@@ -43,13 +52,13 @@ class ClearEnviron(TestCase):
         TestCase.tearDownClass()
 
 
-class CurtsiesPaintingTest(FormatStringTest, ClearEnviron):
+class CurtsiesPaintingTest(ClearEnviron):
     def setUp(self):
         class TestRepl(BaseRepl):
             def _request_refresh(inner_self):
                 pass
 
-        self.repl = TestRepl(config=setup_config())
+        self.repl = TestRepl(setup_config(), cast(CursorAwareWindow, None))
         self.repl.height, self.repl.width = (5, 10)
 
     @property
@@ -58,17 +67,29 @@ class CurtsiesPaintingTest(FormatStringTest, ClearEnviron):
 
     def assert_paint(self, screen, cursor_row_col):
         array, cursor_pos = self.repl.paint()
-        self.assertFSArraysEqual(array, screen)
+        assertFSArraysEqual(array, screen)
         self.assertEqual(cursor_pos, cursor_row_col)
 
     def assert_paint_ignoring_formatting(
         self, screen, cursor_row_col=None, **paint_kwargs
     ):
         array, cursor_pos = self.repl.paint(**paint_kwargs)
-        self.assertFSArraysEqualIgnoringFormatting(array, screen)
+        assertFSArraysEqualIgnoringFormatting(array, screen)
         if cursor_row_col is not None:
             self.assertEqual(cursor_pos, cursor_row_col)
 
+    def process_box_characters(self, screen):
+        if not self.repl.config.unicode_box or not config.supports_box_chars():
+            return [
+                line.replace("┌", "+")
+                .replace("└", "+")
+                .replace("┘", "+")
+                .replace("┐", "+")
+                .replace("─", "-")
+                for line in screen
+            ]
+        return screen
+
 
 class TestCurtsiesPaintingTest(CurtsiesPaintingTest):
     def test_history_is_cleared(self):
@@ -112,22 +133,23 @@ class TestCurtsiesPaintingSimple(CurtsiesPaintingTest):
         self.repl.height, self.repl.width = (5, 32)
         self.repl.current_line = "an"
         self.cursor_offset = 2
-        if config.supports_box_chars():
-            screen = [
+        screen = self.process_box_characters(
+            [
                 ">>> an",
                 "┌──────────────────────────────┐",
                 "│ and  any(                    │",
                 "└──────────────────────────────┘",
                 "Welcome to bpython! Press <F1> f",
             ]
-        else:
-            screen = [
+            if sys.version_info[:2] < (3, 10)
+            else [
                 ">>> an",
-                "+------------------------------+",
-                "| and  any(                    |",
-                "+------------------------------+",
+                "┌──────────────────────────────┐",
+                "│ and    anext( any(           │",
+                "└──────────────────────────────┘",
                 "Welcome to bpython! Press <F1> f",
             ]
+        )
         self.assert_paint_ignoring_formatting(screen, (0, 4))
 
     def test_argspec(self):
@@ -153,7 +175,7 @@ class TestCurtsiesPaintingSimple(CurtsiesPaintingTest):
             + bold(cyan("10"))
             + yellow(")")
         ]
-        self.assertFSArraysEqual(fsarray(array), fsarray(screen))
+        assertFSArraysEqual(fsarray(array), fsarray(screen))
 
     def test_formatted_docstring(self):
         actual = replpainter.formatted_docstring(
@@ -162,7 +184,7 @@ class TestCurtsiesPaintingSimple(CurtsiesPaintingTest):
             config=setup_config(),
         )
         expected = fsarray(["Returns the results", "", "Also has side effects"])
-        self.assertFSArraysEqualIgnoringFormatting(actual, expected)
+        assertFSArraysEqualIgnoringFormatting(actual, expected)
 
     def test_unicode_docstrings(self):
         "A bit of a special case in Python 2"
@@ -175,7 +197,7 @@ class TestCurtsiesPaintingSimple(CurtsiesPaintingTest):
             foo.__doc__, 40, config=setup_config()
         )
         expected = fsarray(["åß∂ƒ"])
-        self.assertFSArraysEqualIgnoringFormatting(actual, expected)
+        assertFSArraysEqualIgnoringFormatting(actual, expected)
 
     def test_nonsense_docstrings(self):
         for docstring in [
@@ -188,7 +210,7 @@ class TestCurtsiesPaintingSimple(CurtsiesPaintingTest):
                     docstring, 40, config=setup_config()
                 )
             except Exception:
-                self.fail("bad docstring caused crash: {!r}".format(docstring))
+                self.fail(f"bad docstring caused crash: {docstring!r}")
 
     def test_weird_boto_docstrings(self):
         # Boto does something like this.
@@ -205,7 +227,7 @@ class TestCurtsiesPaintingSimple(CurtsiesPaintingTest):
         wd = pydoc.getdoc(foo)
         actual = replpainter.formatted_docstring(wd, 40, config=setup_config())
         expected = fsarray(["asdfåß∂ƒ"])
-        self.assertFSArraysEqualIgnoringFormatting(actual, expected)
+        assertFSArraysEqualIgnoringFormatting(actual, expected)
 
     def test_paint_lasts_events(self):
         actual = replpainter.paint_last_events(
@@ -216,7 +238,7 @@ class TestCurtsiesPaintingSimple(CurtsiesPaintingTest):
         else:
             expected = fsarray(["+-+", "|c|", "|b|", "+-+"])
 
-        self.assertFSArraysEqualIgnoringFormatting(actual, expected)
+        assertFSArraysEqualIgnoringFormatting(actual, expected)
 
 
 @contextmanager
@@ -264,7 +286,9 @@ class HigherLevelCurtsiesPaintingTest(CurtsiesPaintingTest):
             def _request_refresh(inner_self):
                 self.refresh()
 
-        self.repl = TestRepl(banner="", config=setup_config())
+        self.repl = TestRepl(
+            setup_config(), cast(CursorAwareWindow, None), banner=""
+        )
         self.repl.height, self.repl.width = (5, 32)
 
     def send_key(self, key):
@@ -634,6 +658,7 @@ class TestCurtsiesRewindRedraw(HigherLevelCurtsiesPaintingTest):
     def test_unhighlight_paren_bugs(self):
         """two previous bugs, parent didn't highlight until next render
         and paren didn't unhighlight until enter"""
+        self.repl.width = 32
         self.assertEqual(self.repl.rl_history.entries, [""])
         self.enter("(")
         self.assertEqual(self.repl.rl_history.entries, [""])
@@ -650,7 +675,8 @@ class TestCurtsiesRewindRedraw(HigherLevelCurtsiesPaintingTest):
             [
                 cyan(">>> ") + on_magenta(bold(red("("))),
                 green("... ") + on_magenta(bold(red(")"))),
-            ]
+            ],
+            width=32,
         )
         self.assert_paint(screen, (1, 5))
 
@@ -660,7 +686,8 @@ class TestCurtsiesRewindRedraw(HigherLevelCurtsiesPaintingTest):
             [
                 cyan(">>> ") + yellow("("),
                 green("... ") + yellow(")") + bold(cyan(" ")),
-            ]
+            ],
+            width=32,
         )
         self.assert_paint(screen, (1, 6))
 
@@ -687,7 +714,7 @@ class TestCurtsiesRewindRedraw(HigherLevelCurtsiesPaintingTest):
 
 
 def completion_target(num_names, chars_in_first_name=1):
-    class Class(object):
+    class Class:
         pass
 
     if chars_in_first_name < 1:
@@ -733,14 +760,16 @@ class TestCurtsiesInfoboxPaint(HigherLevelCurtsiesPaintingTest):
         self.repl.current_line = "abc"
         self.repl.cursor_offset = 3
         self.repl.process_event(".")
-        screen = [
-            ">>> abc.",
-            "+------------------+",
-            "| aaaaaaaaaaaaaaaa |",
-            "| b                |",
-            "| c                |",
-            "+------------------+",
-        ]
+        screen = self.process_box_characters(
+            [
+                ">>> abc.",
+                "┌──────────────────┐",
+                "│ aaaaaaaaaaaaaaaa │",
+                "│ b                │",
+                "│ c                │",
+                "└──────────────────┘",
+            ]
+        )
         self.assert_paint_ignoring_formatting(screen, (0, 8))
 
     def test_fill_screen(self):
@@ -749,23 +778,25 @@ class TestCurtsiesInfoboxPaint(HigherLevelCurtsiesPaintingTest):
         self.repl.current_line = "abc"
         self.repl.cursor_offset = 3
         self.repl.process_event(".")
-        screen = [
-            ">>> abc.",
-            "+------------------+",
-            "| aaaaaaaaaaaaaaaa |",
-            "| b                |",
-            "| c                |",
-            "| d                |",
-            "| e                |",
-            "| f                |",
-            "| g                |",
-            "| h                |",
-            "| i                |",
-            "| j                |",
-            "| k                |",
-            "| l                |",
-            "+------------------+",
-        ]
+        screen = self.process_box_characters(
+            [
+                ">>> abc.",
+                "┌──────────────────┐",
+                "│ aaaaaaaaaaaaaaaa │",
+                "│ b                │",
+                "│ c                │",
+                "│ d                │",
+                "│ e                │",
+                "│ f                │",
+                "│ g                │",
+                "│ h                │",
+                "│ i                │",
+                "│ j                │",
+                "│ k                │",
+                "│ l                │",
+                "└──────────────────┘",
+            ]
+        )
         self.assert_paint_ignoring_formatting(screen, (0, 8))
 
     def test_lower_on_screen(self):
@@ -775,37 +806,41 @@ class TestCurtsiesInfoboxPaint(HigherLevelCurtsiesPaintingTest):
         self.repl.current_line = "abc"
         self.repl.cursor_offset = 3
         self.repl.process_event(".")
-        screen = [
-            ">>> abc.",
-            "+------------------+",
-            "| aaaaaaaaaaaaaaaa |",
-            "| b                |",
-            "| c                |",
-            "| d                |",
-            "| e                |",
-            "| f                |",
-            "| g                |",
-            "| h                |",
-            "| i                |",
-            "| j                |",
-            "| k                |",
-            "| l                |",
-            "+------------------+",
-        ]
+        screen = self.process_box_characters(
+            [
+                ">>> abc.",
+                "┌──────────────────┐",
+                "│ aaaaaaaaaaaaaaaa │",
+                "│ b                │",
+                "│ c                │",
+                "│ d                │",
+                "│ e                │",
+                "│ f                │",
+                "│ g                │",
+                "│ h                │",
+                "│ i                │",
+                "│ j                │",
+                "│ k                │",
+                "│ l                │",
+                "└──────────────────┘",
+            ]
+        )
         # behavior before issue #466
         self.assert_paint_ignoring_formatting(
             screen, try_preserve_history_height=0
         )
         self.assert_paint_ignoring_formatting(screen, min_infobox_height=100)
         # behavior after issue #466
-        screen = [
-            ">>> abc.",
-            "+------------------+",
-            "| aaaaaaaaaaaaaaaa |",
-            "| b                |",
-            "| c                |",
-            "+------------------+",
-        ]
+        screen = self.process_box_characters(
+            [
+                ">>> abc.",
+                "┌──────────────────┐",
+                "│ aaaaaaaaaaaaaaaa │",
+                "│ b                │",
+                "│ c                │",
+                "└──────────────────┘",
+            ]
+        )
         self.assert_paint_ignoring_formatting(screen)
 
     def test_at_bottom_of_screen(self):
@@ -815,35 +850,39 @@ class TestCurtsiesInfoboxPaint(HigherLevelCurtsiesPaintingTest):
         self.repl.current_line = "abc"
         self.repl.cursor_offset = 3
         self.repl.process_event(".")
-        screen = [
-            ">>> abc.",
-            "+------------------+",
-            "| aaaaaaaaaaaaaaaa |",
-            "| b                |",
-            "| c                |",
-            "| d                |",
-            "| e                |",
-            "| f                |",
-            "| g                |",
-            "| h                |",
-            "| i                |",
-            "| j                |",
-            "| k                |",
-            "| l                |",
-            "+------------------+",
-        ]
+        screen = self.process_box_characters(
+            [
+                ">>> abc.",
+                "┌──────────────────┐",
+                "│ aaaaaaaaaaaaaaaa │",
+                "│ b                │",
+                "│ c                │",
+                "│ d                │",
+                "│ e                │",
+                "│ f                │",
+                "│ g                │",
+                "│ h                │",
+                "│ i                │",
+                "│ j                │",
+                "│ k                │",
+                "│ l                │",
+                "└──────────────────┘",
+            ]
+        )
         # behavior before issue #466
         self.assert_paint_ignoring_formatting(
             screen, try_preserve_history_height=0
         )
         self.assert_paint_ignoring_formatting(screen, min_infobox_height=100)
         # behavior after issue #466
-        screen = [
-            ">>> abc.",
-            "+------------------+",
-            "| aaaaaaaaaaaaaaaa |",
-            "| b                |",
-            "| c                |",
-            "+------------------+",
-        ]
+        screen = self.process_box_characters(
+            [
+                ">>> abc.",
+                "┌──────────────────┐",
+                "│ aaaaaaaaaaaaaaaa │",
+                "│ b                │",
+                "│ c                │",
+                "└──────────────────┘",
+            ]
+        )
         self.assert_paint_ignoring_formatting(screen)
